/********************************************************************************/
/* Copyright (C) 2009-2011, Heartland.Data.co.  All Rights Reserved.			*/
/*																				*/
/* Title	: DT10 Driver	[All]												*/
/* Author	: HLDC																*/
/********************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/string.h>
#include <asm/io.h>
#include <linux/ep93xx_gpio.h>

#include "dt10_tpdrv.h"

#define LICENSE "GPL v2"

#define	CS_BIT		0x00000010	/* CS port number */
#define	CLK_BIT		0x00000020	/* CLK port number */
#define	DATA_BIT	0x00000040	/* DATA port number */

#define	CLK_CS_HI	{			\
	int bit;						\
	bit = inl(GPIO_PADR);		\
	bit |= CLK_BIT | CS_BIT;	\
	outl(bit, GPIO_PADR);		\
}

#define	CS_HI	{			\
	int bit;					\
	bit = inl(GPIO_PADR);	\
	bit |= CS_BIT; 			\
	outl(bit, GPIO_PADR);	\
}
#define	CS_LOW	{			\
	int bit;					\
	bit = inl(GPIO_PADR);	\
	bit &= ~CS_BIT; 			\
	outl(bit, GPIO_PADR);	\
}

#define	DATA_OUT(x)	{	\
	int bit;					\
	bit = inl(GPIO_PADR);	\
	bit &= ~CLK_BIT; 		\
	outl(bit, GPIO_PADR);	\
	bit = inl(GPIO_PADR);	\
	if( (x) ){				\
		bit |= DATA_BIT; 	\
	}							\
	else{						\
		bit &= ~DATA_BIT; 	\
	}							\
	outl(bit, GPIO_PADR);	\
	bit = inl(GPIO_PADR);	\
	bit |= CLK_BIT; 			\
	outl(bit, GPIO_PADR);	\
}

/*===============================================================================
	Func	: dt10drvspi_open
	Desc	: This function is executed when opening the device driver.
===============================================================================*/
static int dt10drvspi_open(struct inode *inode, struct file *filp)
{	
	unsigned int bit ;

	/* Port A */
	bit = inl(GPIO_PADDR);
	bit |= DATA_BIT | CLK_BIT | CS_BIT ;
	outl(bit, GPIO_PADDR);

	CLK_CS_HI;

	printk("dt10drvspi: open\n");
	return 0;
}

/*===============================================================================
	Func	: spi_busout
	Desc	: Called by Test Point.
===============================================================================*/
static void spi_busout( unsigned int addr, unsigned int data )
{
	CS_LOW;
	DATA_OUT(data & 0x8000);	// bit15
	DATA_OUT(data & 0x4000);	// bit14
	DATA_OUT(data & 0x2000);	// bit13
	DATA_OUT(data & 0x1000);	// bit12
	DATA_OUT(data & 0x0800);	// bit11
	DATA_OUT(data & 0x0400);	// bit10
	DATA_OUT(data & 0x0200);	// bit9
	DATA_OUT(data & 0x0100);	// bit8
	DATA_OUT(data & 0x0080);	// bit7
	DATA_OUT(data & 0x0040);	// bit6
	DATA_OUT(data & 0x0020);	// bit5
	DATA_OUT(data & 0x0010);	// bit4
	DATA_OUT(data & 0x0008);	// bit3
	DATA_OUT(data & 0x0004);	// bit2
	DATA_OUT(data & 0x0002);	// bit1
	DATA_OUT(data & 0x0001);	// bit0

	if( addr >= 0x00010000 )
	{
		DATA_OUT(addr & 0x800000);	// bit23
		DATA_OUT(addr & 0x400000);	// bit22
		DATA_OUT(addr & 0x200000);	// bit21
		DATA_OUT(addr & 0x100000);	// bit20
		DATA_OUT(addr & 0x080000);	// bit19
		DATA_OUT(addr & 0x040000);	// bit18
		DATA_OUT(addr & 0x020000);	// bit17
		DATA_OUT(addr & 0x010000);	// bit16
	}
	if( addr >= 0x00000100 )
	{
		DATA_OUT(addr & 0x8000);	// bit15
		DATA_OUT(addr & 0x4000);	// bit14
		DATA_OUT(addr & 0x2000);	// bit13
		DATA_OUT(addr & 0x1000);	// bit12
		DATA_OUT(addr & 0x0800);	// bit11
		DATA_OUT(addr & 0x0400);	// bit10
		DATA_OUT(addr & 0x0200);	// bit9
		DATA_OUT(addr & 0x0100);	// bit8
	}
	DATA_OUT(addr & 0x0080);	// bit7
	DATA_OUT(addr & 0x0040);	// bit6
	DATA_OUT(addr & 0x0020);	// bit5
	DATA_OUT(addr & 0x0010);	// bit4
	DATA_OUT(addr & 0x0008);	// bit3
	DATA_OUT(addr & 0x0004);	// bit2
	DATA_OUT(addr & 0x0002);	// bit1
	DATA_OUT(addr & 0x0001);	// bit0
	CS_HI;
}

/*===============================================================================
	Func	: spi_memoryout
	Desc	: Test Point Output Function.
===============================================================================*/
static void spi_memoryout(unsigned char data)
{
	DATA_OUT(data & 0x0080);	// bit7
	DATA_OUT(data & 0x0040);	// bit6
	DATA_OUT(data & 0x0020);	// bit5
	DATA_OUT(data & 0x0010);	// bit4
	DATA_OUT(data & 0x0008);	// bit3
	DATA_OUT(data & 0x0004);	// bit2
	DATA_OUT(data & 0x0002);	// bit1
	DATA_OUT(data & 0x0001);	// bit0
}

/*===============================================================================
	Func	: spi_busoutfast
	Desc	:  Test Point Output Function.
===============================================================================*/
static void spi_busoutfast(unsigned int data, unsigned int bit)
{
	CS_LOW;
	if (bit >= 10) DATA_OUT(data & 0x0200);	// bit9
	if (bit >= 9) DATA_OUT(data & 0x0100);	// bit8
	if (bit >= 8) DATA_OUT(data & 0x0080);	// bit7
	if (bit >= 7) DATA_OUT(data & 0x0040);	// bit6
	if (bit >= 6) DATA_OUT(data & 0x0020);	// bit5
	if (bit >= 5) DATA_OUT(data & 0x0010);	// bit4
	if (bit >= 4) DATA_OUT(data & 0x0008);	// bit3
	if (bit >= 3) DATA_OUT(data & 0x0004);	// bit2
	if (bit >= 2) DATA_OUT(data & 0x0002);	// bit1
	DATA_OUT(data & 0x0001);					// bit0
	CS_HI;
}

/*===============================================================================
	Func	: spi_eventout
	Desc	:Test Point Output Function.
===============================================================================*/
static void spi_eventout(unsigned int addr, unsigned int data)
{
	CS_LOW;
	DATA_OUT(addr & 0x0002);	// bit1
	DATA_OUT(addr & 0x0001);	// bit0
	CS_HI;
	CS_LOW;
	spi_memoryout((data));
	spi_memoryout((data >> 8));
	spi_memoryout(0x00);
	spi_memoryout(0x00);
	CS_HI;
}

/*===============================================================================
	Func	: dt10drvspi_write
	Desc	: This function is executed upon write instruction.
===============================================================================*/
static int dt10drvspi_write(struct file *filp, char *buff, size_t count, loff_t *pos)
{
	dt10_tp *tp;
	int	i;
	
	tp = (dt10_tp *)buff ;
	switch (tp->fast_mode) {
	case 0:					/* Normal Mode */
		spi_busout(tp->addr, tp->data);
		break;
	case 1:					/* Fast Mode */
		spi_busoutfast(tp->addr, tp->data);
		break;
	case 2:					/* Fast Mode Event Trigger */
		spi_eventout(tp->addr, tp->data);
		break;
	default:
		break;
	}
	if( tp->size != 0 ){
		CS_LOW;
		for (i = 0; i < tp->size; i++) {
			spi_memoryout(tp->p[i]);
		}
		CS_HI;
	}
	return count ;

}

/*===============================================================================
	Func	: dt10drvspi_release
	Desc	: This function is executed when closing the device driver.
===============================================================================*/
static int dt10drvspi_release(struct inode *inode, struct file *filp)
{
	printk("dt10drvspi: release\n");
	return 0;
}

/*===============================================================================
	Func	: file_operations driver_fops
===============================================================================*/
static struct file_operations driver_fops = {
	.write = dt10drvspi_write,
	.open = dt10drvspi_open,
	.release = dt10drvspi_release,
};

/*===============================================================================
	Func	: miscdevice misc
===============================================================================*/
static struct miscdevice misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "dt10drvspi",
	.fops = &driver_fops,
};
/*===============================================================================
	Func	: init_module_spi
	Desc	: This function is executed when installing the device driver.
===============================================================================*/
int __init init_module_spi(void)
{
	int ret;
	
	ret = misc_register(&misc);
	if (ret < 0) {
		printk(KERN_ERR "dt10drvspi: misc_register() failed\n");
	}
	printk("dt10drvspi: Success init_module.\n");

	return 0;
}

/*===============================================================================
	Func	: cleanup_module_spi
	Desc	: This function is executed when uninstalling the device driver.
===============================================================================*/
void __exit cleanup_module_spi(void)
{
	misc_deregister(&misc);
	printk("dt10drvspi: cleanup_module_spi\n");
}

module_init(init_module_spi);
module_exit(cleanup_module_spi);

MODULE_LICENSE(LICENSE);

